import St from 'gi://St'
import Meta from 'gi://Meta'
import Clutter from 'gi://Clutter'
import Shell from 'gi://Shell'

import * as DND from 'resource:///org/gnome/shell/ui/dnd.js'
import * as Main from 'resource:///org/gnome/shell/ui/main.js'
import * as PopupMenu from 'resource:///org/gnome/shell/ui/popupMenu.js'
import * as AppFavorites from 'resource:///org/gnome/shell/ui/appFavorites.js'
import * as BoxPointer from 'resource:///org/gnome/shell/ui/boxpointer.js'

import {Extension} from 'resource:///org/gnome/shell/extensions/extension.js'

const CYCLING_TIMEOUT_MS = 1000
const APPLICATION_ICON_SIZE = 32

export default class VerticalWindowListExtension extends Extension {
    enable() {
        let ref, ref1
        this.settings = this.getSettings()
        this.settings.set_string('windows', '{}')
        this.appSystem = Shell.AppSystem.get_default()
        this.monitor = Main.layoutManager.primaryMonitor
        this.windowPanel = new St.Widget({
            style_class: 'windows-panel',
            reactive: true,
            track_hover: true,
            visible: true,
            layout_manager: new Clutter.BoxLayout({
                orientation: Clutter.Orientation.VERTICAL
            }),
            height: ((ref = this.monitor) != null ? ref.height : void 0) || 1000,
            width: this.settings.get_int('width')
        })
        this.windowPanel.set_position(((ref1 = this.monitor) != null ? ref1.x : void 0) || 0, 30)
        this.windowList = new St.Widget({
            style_class: 'windows-list',
            reactive: true,
            layout_manager: new Clutter.BoxLayout({
                homogeneous: true,
                orientation: Clutter.Orientation.VERTICAL
            }),
            y_align: Clutter.ActorAlign.START,
            x_expand: false,
            y_expand: true
        })

        // non-transparent text and icons
        this.windowList.set_opacity_override(255)
        this.favoritesList = null
        this.windowTracker = Shell.WindowTracker.get_default()
        this.windowsSaved = JSON.parse(this.settings.get_string('windows') || '{}')

        // cycle windows keybingings
        this.cycling = 0
        Main.wm.addKeybinding('cycle-up', this.settings, Meta.KeyBindingFlags.NONE, Shell.ActionMode.NORMAL, this.cycleWindow.bind(this, true))
        Main.wm.addKeybinding('cycle-down', this.settings, Meta.KeyBindingFlags.NONE, Shell.ActionMode.NORMAL, this.cycleWindow.bind(this, false))
        this.windowPanel.connect('destroy', () => {
            Main.wm.removeKeybinding('cycle-up')
            Main.wm.removeKeybinding('cycle-down')
        })

        {
            let signal = this.settings.connect('changed::panel-opacity', this.opacityChanged.bind(this))
            this.windowPanel.connect('destroy', () => { this.settings.disconnect(signal) })
        }

        {
            let signal = this.settings.connect('changed::width', this.widthChanged.bind(this))
            this.windowPanel.connect('destroy', () => { this.settings.disconnect(signal) })
        }

        {
            let signal = this.windowPanel.connect('button-release-event', this.panelClicked.bind(this))
            this.windowPanel.connect('destroy', () => { this.windowPanel.disconnect(signal) })
        }

        {
            let signal = global.workspace_manager.connect('active-workspace-changed', this.addWorkspaceWindows.bind(this))
            this.windowPanel.connect('destroy', () => { global.workspace_manager.disconnect(signal) })
        }

        {
            let signal = Main.layoutManager.connect('monitors-changed', () => {
                let ref2, ref3
                this.monitor = Main.layoutManager.primaryMonitor
                this.windowPanel.set_position(((ref2 = this.monitor) != null ? ref2.x : void 0) || 0, 30)
                this.windowPanel.set_height(((ref3 = this.monitor) != null ? ref3.height : void 0) || 1000)
                Main.layoutManager.removeChrome(this.windowPanel)
                return Main.layoutManager.addChrome(this.windowPanel, {
                    affectsStruts: true,
                    trackFullscreen: true
                })
            })
            this.windowPanel.connect('destroy', () => { Main.layoutManager.disconnect(signal) })
        }

        this.windowPanel.add_child(this.windowList)
        Main.layoutManager.addChrome(this.windowPanel, {
            affectsStruts: true,
            trackFullscreen: true
        })

        this.addWorkspaceWindows()
        this.opacityChanged()
    }

    addWorkspaceWindows() {
        let button, i, j, k, len, len1, len2, metaWindow, ref, ref1, w, windowPresent, windowSaved, windowsPresent,
            workspace

        workspace = global.workspace_manager.get_active_workspace()
        this.saveWorkspaceWindows()
        ref = this.windowList.get_children()
        for (i = 0, len = ref.length; i < len; i++) {
            button = ref[i]
            button.destroy()
        }
        windowsPresent = []
        ref1 = global.display.get_tab_list(Meta.TabList.NORMAL, workspace)
        for (j = 0, len1 = ref1.length; j < len1; j++) {
            metaWindow = ref1[j]
            windowPresent = {
                order: metaWindow.get_stable_sequence(),
                windowId: metaWindow.get_id(),
                metaWindow: metaWindow,
                workspaceIndex: workspace.index()
            }
            windowSaved = this.windowsSaved[windowPresent.windowId]
            if (windowSaved) {
                windowPresent.order = windowSaved.order
                windowPresent.customTitle = windowSaved.customTitle
            }
            windowsPresent.push(windowPresent)
        }
        windowsPresent.sort(function (a, b) {
            return 0 + (a.order >= b.order) - (a.order <= b.order)
        })
        for (k = 0, len2 = windowsPresent.length; k < len2; k++) {
            w = windowsPresent[k]
            this.add_window(workspace, w.metaWindow, w.customTitle)
        }

        if (workspace._window_added == null) {
            workspace._window_added = workspace.connect_after('window-added', this.add_window.bind(this))
        }

        if (workspace._window_removed == null) {
            workspace._window_removed = workspace.connect_after('window-removed', this.remove_window.bind(this))
        }
    }

    cycleWindow(up, display, window, binding) {
        let b, buttons

        // cycle up or down the list starting from previous Window
        buttons = this.windowList.get_children()
            let [activeWindow, ix] = ((function () {
            let i, len, results
            results = []
            for (ix = i = 0, len = buttons.length; i < len; ix = ++i) {
                b = buttons[ix]
                if (b.has_style_class_name('focused')) {
                    results.push([b.metaWindow, ix])
                }
            }
            return results
        })())[0] || []
        if (activeWindow == null) {
            return
        }
        if (Date.now() - this.cycling > CYCLING_TIMEOUT_MS) {
            this.activateWindow(this.previousWindow(activeWindow))
            this.cycling = Date.now()
            return
        }
        if (up) {
            ix = (ix === buttons.length - 1) ? 0 : ix + 1
        } else {
            ix = ix === 0 ? buttons.length - 1 : ix - 1
        }
        this.activateWindow(buttons[ix].metaWindow)
        this.cycling = Date.now()
    }

    panelClicked(panel, event) {
        let ref
        if (event.get_button() === 1) {
            if ((ref = this.favoritesList) != null ? ref.isOpen : void 0) {
                this.hideFavorites()
            } else {
                this.showFavorites(event)
            }
        }
        if (event.get_button() === 2) {
            this.showSettings()
        }
        return Clutter.EVENT_STOP
    }

    showSettings() {
        this.openPreferences()
    }

    showFavorites(event) {
        let actor, apps, btn, i, item, j, len, len1, ref, x, y

        this.hideFavorites()
        ref = this.windowList.get_children()
        // apply custom title if it was in edit mode
        for (i = 0, len = ref.length; i < len; i++) {
            btn = ref[i]
            if (btn.custom_title.visible) {
                this.updateTitle(btn.metaWindow, btn)
            }
        }
        [x, y] = event.get_coords()
        actor = new St.Widget({
            width: 0,
            height: 0,
            opacity: 0
        })
        this.favoritesList = new PopupMenu.PopupMenu(actor, 0, St.Side.TOP)
        apps = AppFavorites.getAppFavorites().getFavorites()
        if (!apps) {
            return
        }
        for (j = 0, len1 = apps.length; j < len1; j++) {
            let app = apps[j]
            item = new PopupMenu.PopupImageMenuItem(app.get_name(), app.create_icon_texture(APPLICATION_ICON_SIZE).get_gicon())
            // PopupMenu closes and disconnects signal automatically
            item.connect('activate', () => { app.open_new_window(-1) })

            this.favoritesList.addMenuItem(item)
        }
        Main.layoutManager.addTopChrome(this.favoritesList.actor)
        this.favoritesList.open(BoxPointer.PopupAnimation.NONE)
        // move favorites higher when it does not fit vertically and click was in lower half of monitor
        if (y > this.monitor.height / 2 && y + this.favoritesList.actor.get_height() > this.monitor.height) {
            y = this.monitor.height - this.favoritesList.actor.get_height() - 10
        }
        this.favoritesList.actor.set_x(x + 10)
        this.favoritesList.actor.set_y(y + 10)
    }

    hideFavorites() {
        let ref
        if ((ref = this.favoritesList) != null ? ref.isOpen : void 0) {
            this.favoritesList.destroy()
            this.favoritesList = null
        }
    }

    widthChanged() {
        this.windowPanel.width = this.settings.get_int('width')
    }

    opacityChanged() {
        this.windowPanel.set_opacity(this.settings.get_int('panel-opacity'))
    }

    add_window(workspace, metaWindow, customTitle = '') {
        let button, texture_cache

        if (workspace == null) {
            workspace = global.workspace_manager.get_active_workspace()
        }
        if (global.display.get_tab_list(Meta.TabList.NORMAL, workspace).indexOf(metaWindow) < 0) {
            return
        }

        this.remove_window(workspace, metaWindow)
        button = new St.Widget({
            style_class: 'window-button',
            can_focus: true,
            x_expand: true,
            x_align: Clutter.ActorAlign.FILL,
            reactive: true,
            layout_manager: new Clutter.BoxLayout({
                orientation: Clutter.Orientation.HORIZONTAL
            })
        })
        button.metaWindow = metaWindow
        button.title = new St.Label({
            x_expand: true,
            y_align: Clutter.ActorAlign.CENTER,
            text: '',
            style_class: 'window-button-text'
        })
        button.title.clutter_text.single_line_mode = true
        button.icon = new St.Bin({
            x_expand: false,
            style_class: 'window-button-icon'
        })
        button.custom_title = new St.Entry({
            y_align: Clutter.ActorAlign.CENTER,
            text: customTitle,
            style_class: 'window-button-editor',
            visible: false
        })
        button.add_child(button.icon)
        button.add_child(button.title)
        button.add_child(button.custom_title)
        this.windowList.add_child(button)
        this.updateTitle(metaWindow, button)
        this.updateIcon(metaWindow, button)
        {
            // apply custom title
            let signal = button.custom_title.connect('key-release-event', (custom_title, event) => {
                let ref, ref1
                if ((ref = event.get_key_code()) === 36 || ref === 104) { // Enter
                    this.updateTitle(metaWindow, button)
                }
                if ((ref1 = event.get_key_code()) === 9) { // Escape
                    button.custom_title.text = ''
                    this.updateTitle(metaWindow, button)
                }
            })
            button.connect('destroy', () => { button.custom_title.disconnect(signal) })
        }

        // windows reordering via drag and drop
        button.draggable = DND.makeDraggable(button, {
            dragActorOpacity: 127
        })
        button.draggable.connect('drag-end', this.dragEnd.bind(this, button))
        button.draggable.connect('drag-begin', function () {
            let dragActor = new St.Button({
                style_class: 'drag-button',
                label: '' + metaWindow.title
            })

            let [sizeW, sizeH] = button.get_size()
            dragActor.set_size(sizeW, sizeH)

            let [posX, posY] = button.get_transformed_position()
            dragActor.set_position(posX, posY)
            dragActor.getDragActor = function () {
                return dragActor
            }
            button._delegate = dragActor

        })

        {
            let signal = button.connect('button-release-event', this.buttonClicked.bind(this))
            button.connect('destroy', () => { button.disconnect(signal) })
        }

        {
            let texture_cache = St.TextureCache.get_default()
            let signal = texture_cache.connect_after('icon-theme-changed', this.updateIcon.bind(this, metaWindow, button))
            button.connect('destroy', () => { texture_cache.disconnect(signal) })
        }

        {
            let signal = metaWindow.connect_after('notify::wm-class', this.updateIcon.bind(this, metaWindow, button))
            button.connect('destroy', () => { metaWindow.disconnect(signal) })
        }

        {
            let signal = this.appSystem.connect('app-state-changed', this.updateIcon.bind(this, metaWindow, button))
            button.connect('destroy', () => { this.appSystem.disconnect(signal) })
        }


        {
            let signal = metaWindow.connect_after('notify::gtk-application-id', this.updateIcon.bind(this, metaWindow, button))
            button.connect('destroy', () => { metaWindow.disconnect(signal) })
        }

        {
            let signal = this.settings.connect('changed::icon-size', this.updateIcon.bind(this, metaWindow, button))
            button.connect('destroy', () => { this.settings.disconnect(signal) })
        }

        {
            let signal = metaWindow.connect_after('notify::title', this.updateTitle.bind(this, metaWindow, button))
            button.connect('destroy', function () { metaWindow.disconnect(signal) })
        }

        {
            let signal = this.appSystem.connect('app-state-changed', this.updateTitle.bind(this, metaWindow, button))
            button.connect('destroy', () => { this.appSystem.disconnect(signal) })
        }

        {
            let signal = global.display.connect('notify::focus-window', this.focusWindow.bind(this, button))
            button.connect('destroy', () => { global.display.disconnect(signal) })
        }
    }

    dragEnd(button) {
        let b, bRect, i, ix, len, ref

        let dropRect = new Meta.Rectangle({
            x: button.draggable._dragX,
            y: button.draggable._dragY,
            width: 1,
            height: 1
        })

        ref = this.windowList.get_children()
        for (ix = i = 0, len = ref.length; i < len; ix = ++i) {
            b = ref[ix]
            bRect = new Meta.Rectangle({
                x: b.get_transformed_position()[0],
                y: b.get_transformed_position()[1],
                width: b.get_transformed_size()[0],
                height: b.get_transformed_size()[1]
            })
            if (b === button) {
                continue
            }
            if (bRect.contains_rect(dropRect)) {
                this.windowList.remove_child(button)
                this.windowList.insert_child_at_index(button, ix)
            }
        }

        // drag under last window
        if (dropRect.y > bRect.y) {
            this.windowList.remove_child(button)
            return this.windowList.add_child(button)
        }
    }

    remove_window(workspace, metaWindow) {
        let button, i, len, ref

        ref = this.windowList.get_children()
        for (i = 0, len = ref.length; i < len; i++) {
            button = ref[i]
            if (button.metaWindow === metaWindow) {
                button.destroy()
            }
        }
    }

    focusWindow(button) {
        if (global.display.focus_window === button.metaWindow) {
            button.add_style_class_name('focused')
        } else {
            button.remove_style_class_name('focused')
        }
    }

    updateTitle(metaWindow, button) {
        button.title.set_text(button.custom_title.text || ('' + metaWindow.title))
        button.custom_title.visible = false
        button.title.visible = true
    }

    app(metaWindow) {
        return this.windowTracker.get_window_app(metaWindow)
    }

    updateIcon(metaWindow, button) {
        let app, base
        app = this.app(metaWindow)
        if (app) {
            button.icon.child = app.create_icon_texture(this.settings.get_int('icon-size'))
        } else {
            base = button.icon
            if (base.child == null) {
                base.child = new St.Icon({
                    icon_name: 'icon-missing',
                    icon_size: this.settings.get_int('icon-size')
                })
            }
        }
    }

    buttonClicked(button, event) {
        let  btn, button_number, i, len, metaWindow, prevWindow, ref

        this.hideFavorites()
        ref = this.windowList.get_children()
        // apply custom title if it was in edit mode
        for (i = 0, len = ref.length; i < len; i++) {
            btn = ref[i]
            if (btn.custom_title.visible) {
                this.updateTitle(metaWindow, btn)
            }
        }

        // activate previous window if already was active
        button_number = event.get_button()
        // click through when close to left border
        if (event.get_coords()[0] < 10) {
            this.showFavorites(event)
            return Clutter.EVENT_STOP
        }

        // activate window
        metaWindow = button.metaWindow
        if (button_number === 1) {
            if (metaWindow.appears_focused) {
                prevWindow = this.previousWindow(metaWindow)
                this.activateWindow(prevWindow)
            } else {
                this.activateWindow(metaWindow)
            }
            return Clutter.EVENT_STOP
        }

        // edit custom title
        if (button_number === 2) {
            button.title.visible = false
            button.custom_title.visible = true
            button.custom_title.x_expand = true
            button.custom_title.grab_key_focus()
            return Clutter.EVENT_STOP
        }

        // close window on right click
        if (button_number === 3) {
            this.activateWindow(metaWindow)
            metaWindow.delete(global.get_current_time())
            return Clutter.EVENT_STOP
        }
    }

    activateWindow(metaWindow) {
        metaWindow.activate(global.get_current_time())
        this.cycling = 0
    }

    previousWindow(metaWindow) {
        global.display.get_tab_next(Meta.TabList.NORMAL, metaWindow.get_workspace(), metaWindow, false)
    }

    saveWorkspaceWindows() {
        let button, i, ix, len, metaWindow, ref, results

        ref = this.windowList.get_children()
        for (ix = i = 0, len = ref.length; i < len; ix = ++i) {
            button = ref[ix]
            metaWindow = button.metaWindow
            this.windowsSaved[metaWindow.get_id()] = {
                order: ix + 1,
                customTitle: button.custom_title.text,
                workspaceIndex: metaWindow.get_workspace().index()
            }
        }
    }

    disable() {
        this.saveWorkspaceWindows()
        this.settings.set_string('windows', JSON.stringify(this.windowsSaved))
        this.windowPanel.destroy()


        this.windowPanel = null
        this.settings = null
        this.appSystem = null
        this.monitor = null
        this.windowList = null
        this.favoritesList = null
        this.windowTracker = null
        this.windowsSaved = null
        this.cycling = null
    }

}
